import { evaluate, reset, shift } from "./deps.js";
import { createFuture } from "./future.js";
import { lazy } from "./lazy.js";
export function run(block) {
    let newFrame;
    evaluate(function* () {
        newFrame = yield* createFrame();
        yield* newFrame.enter(block);
    });
    //@ts-expect-error not sure how to my TS recognize this
    let { frame } = newFrame;
    return createTask(frame);
}
function createTask(frame) {
    let { future, resolve, reject } = createFuture();
    evaluate(function* () {
        let { outcome } = yield* frame;
        if (outcome.type === "resolved") {
            resolve(outcome.value);
        }
        else {
            reject(outcome.error);
        }
    });
    return {
        ...future,
        halt() {
            let { future, resolve, reject } = createFuture();
            evaluate(function* () {
                let result = yield* frame.destroy();
                if (result.type === "resolved") {
                    resolve(result.value);
                }
                else {
                    reject(result.error);
                }
            });
            return future;
        },
    };
}
let ids = 0;
function* createFrame(parent) {
    let id = parent ? `${parent.id}.${parent.ids++}` : `${ids++}`;
    let context = parent ? Object.create(parent.context) : {};
    let listeners = [];
    let final;
    let resources = new Set();
    let controller = new AbortController();
    let { signal } = controller;
    let frame = {
        id,
        ids: 1,
        context,
        resources,
        state: { type: "running", current: { type: "resolved", value: void 0 } },
        *[Symbol.iterator]() {
            if (final) {
                return final;
            }
            else {
                return yield* shift(function* (k) {
                    listeners.push(k);
                });
            }
        },
        *destroy() {
            controller.abort();
            let { destruction: result } = yield* frame;
            return result;
        },
    };
    let enter = yield* reset(function* () {
        let block = yield* shift(function* (k) {
            return function* (block) {
                k(block);
                return yield* frame;
            };
        });
        let iterator = lazy(() => block()[Symbol.iterator]());
        let exitState = yield* reduce({
            frame,
            iterator,
            start: $next(undefined),
            signal,
        });
        // exit state has now been determined, so time to begin
        // shutdown process.
        // First cleanup anything that we might have been yielding to. This
        // will happen when exitState is a termination or a resource failure.
        let cleanup = { type: "resolved" };
        if (exitState.type !== "result" && exitState.state.type === "yielding") {
            cleanup = yield* exitState.state.to.destroy();
        }
        // Now we run the iterator to completion, no matter what.
        // This cannot be aborted. We may want to warn if we see
        // a suspend instruction in this reduction.
        let exhaustion = yield* reduce({
            frame,
            iterator: iterator,
            start: $abort(),
        });
        // The last bit of cleanup is to terminate all resources since they
        // may have been needed by the scope of the iterator.
        // TODO:
        let deallocation = { type: "resolved" };
        for (let resource of [...resources].reverse()) {
            let dealloc = yield* resource.destroy();
            if (dealloc.type === "rejected") {
                deallocation = dealloc;
            }
        }
        let destruction = { type: "resolved" };
        if (deallocation.type === "rejected") {
            destruction = deallocation;
        }
        else if (exhaustion.type === "result") {
            destruction = exhaustion.result;
        }
        else if (exhaustion.type === "failure") {
            destruction = { type: "rejected", error: exhaustion.error };
        }
        else if (cleanup.type === "rejected") {
            destruction = cleanup;
        }
        else {
            destruction = { type: "resolved" };
        }
        let outcome = {
            type: "rejected",
            error: new Error("outcome unknown"),
        };
        if (destruction.type === "rejected") {
            outcome = destruction;
        }
        else if (exitState.type === "result") {
            outcome = exitState.result;
        }
        else if (exitState.type === "termination") {
            outcome = { type: "rejected", error: new Error("halted") };
        }
        else {
            outcome = { type: "rejected", error: exitState.error };
        }
        //determine final Outcome, and then notify that we're done
        final = { outcome, exit: exitState, destruction };
        for (let listener of listeners) {
            listener(final);
        }
    });
    return { frame, enter };
}
function* reduce(options) {
    let { iterator, frame, signal } = options;
    let { resources } = frame;
    return yield* shift(function* (exit) {
        if (signal) {
            let listener = () => exit({ type: "termination", state: frame.state });
            signal.addEventListener("abort", listener, {
                once: true,
            });
        }
        let getNext = options.start;
        while (!signal || !signal.aborted) {
            let next;
            try {
                next = getNext(iterator());
            }
            catch (error) {
                exit({
                    type: "result",
                    result: { type: "rejected", error },
                });
                break;
            }
            if (next.done) {
                exit({
                    type: "result",
                    result: { type: "resolved", value: next.value },
                });
                break;
            }
            else {
                let instruction = next.value;
                if (instruction.type === "suspend") {
                    let { then } = instruction;
                    frame.state = { type: "suspended" };
                    yield* shift(function* () {
                        then && then();
                    });
                }
                else if (instruction.type === "action") {
                    let { operation } = instruction;
                    let yieldingTo = yield* createFrame(frame);
                    frame.state = { type: "yielding", to: yieldingTo.frame };
                    let result = yield* shift(function* instruction(k) {
                        let $return = (result) => {
                            evaluate(function* () {
                                let termination = yield* yieldingTo.frame.destroy();
                                if (termination.type === "resolved") {
                                    k(result);
                                }
                                else {
                                    k(termination);
                                }
                            });
                        };
                        let resolve = (value) => $return({ type: "resolved", value });
                        let reject = (error) => $return({ type: "rejected", error });
                        let final = yield* yieldingTo.enter(() => operation(resolve, reject));
                        if (final.exit.type === "result" &&
                            final.exit.result.type === "resolved") {
                            k({
                                type: "rejected",
                                error: new Error("reached the end of an action, but resolve() or reject() were never called"),
                            });
                        }
                    });
                    frame.state = { type: "running", current: result };
                    getNext = result.type === "resolved"
                        ? $next(result.value)
                        : $throw(result.error);
                }
                else if (instruction.type === "resource") {
                    let { operation } = instruction;
                    let resource = yield* createFrame(frame);
                    frame.state = { type: "yielding", to: resource.frame };
                    let result = yield* shift(function* (k) {
                        let provisioned = false;
                        let provide = (value) => {
                            provisioned = true;
                            let then = () => {
                                k({ type: "resolved", value });
                            };
                            return [{ type: "suspend", then }];
                        };
                        yield* reset(function* () {
                            let { outcome } = yield* resource.enter(() => operation(provide));
                            if (outcome.type === "rejected") {
                                if (provisioned) {
                                    let { state } = frame;
                                    exit({ type: "failure", state, error: outcome.error });
                                }
                                else {
                                    k(outcome);
                                }
                            }
                        });
                    });
                    frame.state = { type: "running", current: result };
                    resources.add(resource.frame);
                    getNext = result.type === "resolved"
                        ? $next(result.value)
                        : $throw(result.error);
                }
                else if (instruction.type === "getframe") {
                    frame.state = {
                        type: "running",
                        current: { type: "resolved", value: frame },
                    };
                    getNext = $next(frame);
                }
                else if (instruction.type === "spawn") {
                    let { operation } = instruction;
                    let { enter, frame: child } = yield* createFrame(frame);
                    frame.resources.add(child);
                    yield* reset(function* () {
                        let final = yield* enter(operation);
                        let { outcome, destruction } = final;
                        frame.resources.delete(child);
                        let successfullyTerminated = final.exit.type === "termination" &&
                            destruction.type === "resolved";
                        if (outcome.type === "rejected" && !successfullyTerminated) {
                            let { state } = frame;
                            let { error } = outcome;
                            exit({ type: "failure", state, error });
                        }
                    });
                    let task = createTask(child);
                    frame.state = {
                        type: "running",
                        current: { type: "resolved", value: task },
                    };
                    getNext = $next(createTask(child));
                }
            }
        }
    });
}
// deno-lint-ignore no-explicit-any
const $next = (value) => (i) => i.next(value);
const $throw = (error) => (i) => {
    if (i.throw) {
        return i.throw(error);
    }
    else {
        throw error;
    }
};
const $abort = (value) => (i) => {
    if (i.return) {
        return i.return(value);
    }
    else {
        return { done: true, value };
    }
};
//# sourceMappingURL=run.js.map