"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.yaml2ps = exports.matchRefs = exports.ps2js = exports.js2ps = void 0;
/**
 * Convert a JavaScript value into a PlatformScript value
 */
function js2ps(value) {
    let type = typeof value;
    if (type === "number") {
        return { type, value: value };
    }
    else if (type === "boolean") {
        return { type, value: value };
    }
    else if (type === "string") {
        return { type, value: value, holes: [] };
    }
    else if (type === "object") {
        if (Array.isArray(value)) {
            return { type: "list", value: value.map(js2ps) };
        }
        else {
            if (typeof value === "undefined") {
                throw new Error(`'undefined' cannot be converted into a PlatformScript value`);
            }
            let record = value;
            return {
                type: "map",
                value: new Map(Object.keys(record).flatMap((key) => {
                    if (typeof key === "string") {
                        let pskey = { type: "string", value: key, holes: [] };
                        let entry = [pskey, js2ps(record[key])];
                        return [entry];
                    }
                    else {
                        return [];
                    }
                })),
            };
        }
    }
    else if (type === "function") {
        let fn = value;
        return {
            type: "fn",
            *value({ arg }) {
                return js2ps(fn(ps2js(arg)));
            },
        };
    }
    else {
        throw new Error(`cannot convert JavaScript value: '${value}' into PlatformScript`);
    }
}
exports.js2ps = js2ps;
/**
 * Convert a PlatformScript value into a JavaScript value
 */
function ps2js(value) {
    switch (value.type) {
        case "list":
            return value.value.map(ps2js);
        case "map":
            return [...value.value.entries()].reduce((obj, [k, v]) => {
                return Object.assign(obj, { [k.value.toString()]: v });
            }, {});
        case "fn":
            throw new Error("TODO: convert PlatformScript fn into callable JS fn");
        case "ref":
            throw new Error("TODO: de-reference YS references when converting to JS");
        default:
            return value.value;
    }
}
exports.ps2js = ps2js;
function matchRefs(value) {
    let valueIdx = 0;
    let pathIdx = 1;
    let keyIdx = 2;
    let regex = /\$(([\w\d-]+)(\.[\w\d-]+)*)/gmd;
    let i = value.matchAll(regex);
    let holes = [];
    for (let next = i.next(); !next.done; next = i.next()) {
        let match = next.value;
        let ref = {
            type: "ref",
            value: match[valueIdx],
            key: match[keyIdx],
            path: match[pathIdx].split("."),
        };
        //@ts-expect-error RegExpMatchArray#indices not yet in the default TS lib
        holes.push({ ref, range: match.indices[valueIdx] });
    }
    return {
        total: holes.length === 1 && holes[0].ref.value === value,
        holes,
    };
}
exports.matchRefs = matchRefs;
/**
 * Convert source YAML into a PlatformScript Literal
 */
function yaml2ps(node) {
    if (node.kind === 0) {
        let scalar = node;
        if (scalar.singleQuoted) {
            return {
                type: "string",
                node,
                value: scalar.value,
                holes: [],
            };
        }
        else if (scalar.doubleQuoted) {
            return {
                type: "string",
                node,
                value: scalar.value,
                holes: matchRefs(scalar.value).holes,
            };
        }
        else if (typeof scalar.valueObject !== "undefined") {
            let value = scalar.valueObject;
            let type = typeof scalar.valueObject;
            if (!["number", "boolean"].includes(type)) {
                throw new Error(`unknown scalar type: ${type}`);
            }
            return { type: type, value, node };
        }
        else {
            let match = matchRefs(scalar.value);
            if (match.total) {
                return {
                    ...match.holes[0].ref,
                    node,
                };
            }
            else {
                return {
                    type: "string",
                    node,
                    value: scalar.value,
                    holes: match.holes,
                };
            }
        }
    }
    else if (node.kind === 2) {
        let mappings = node.mappings ?? [];
        let [first] = mappings;
        if (!first) {
            return { type: "map", value: new Map(), node };
        }
        else {
            let fnmatch = first.key.value.match(/^\$\s*\((.*)\)\s*$/);
            if (fnmatch) {
                let params = fnmatch[1];
                let body = yaml2ps(first.value);
                return {
                    type: "fn",
                    node,
                    *value({ arg, env }) {
                        let key = {
                            type: "string",
                            node: first.key,
                            value: params,
                            holes: [],
                        };
                        let binding = {
                            type: "map",
                            value: new Map([[key, yield* env.eval(arg)]]),
                        };
                        return yield* env.eval(body, binding);
                    },
                };
            }
            else {
                return {
                    type: "map",
                    node,
                    value: new Map(mappings.map((m) => {
                        let key = yaml2ps(m.key);
                        let value = yaml2ps(m.value);
                        return [key, value];
                    })),
                };
            }
        }
    }
    else if (node.kind === 3) {
        let list = node;
        return {
            type: "list",
            node,
            value: list.items.map(yaml2ps),
        };
    }
    else {
        console.dir({ node }, { depth: 10 });
        throw new Error(`unknown YAMLNode of kind ${node.kind}`);
    }
}
exports.yaml2ps = yaml2ps;
//# sourceMappingURL=convert.js.map